package Data;

use strict;
use warnings;

=head1 data

Accumulates and stores log data. Provides necessary routines to
access database. Allows to store data under variable-length keys.

Prior to use, should be setuped, providing all necessary
keyfields combinations and aggregation functions associated with
them.

Supports aggregations SUM(), MAX(), MIN(), AVG(), COUNT(*),
COUNT(field) with the same meaning as in SQL.
Aggregations get GROUPED BY the same keys as they stored under.

Theoretically, allows to store additional data under names, other
then keyfields or functions - using common access scheme.

Keys, used to access data are variable-length hashes in form of
(keyname=>keyvalue). Data, accumulated by aggregations,
referenced as FUN=>ARG, where FUN is one of supported functions
and ARG is a datafield. These keys transformed into referenced to
data by function Data::ptr().

Due to internal structure, incomplete keys (not containing
aggregation key, or containing set of keynames, not configured at
startup) may point to references to hashes, which may be
traversed to get deeper in database tree. Actually there's
Data::grep() to do it better and independant on future internal
structure changes.

Actual internal access scheme currently is:
 $data->{key1name}{key1value}...{keyNname}{keyNvalue}{FUN}{ARG}

Relative order of keys is (randomly) fixed at startup in
@Data::keyfields, and all possible combinations of keys in path
are pregenerated at initialization and stored in @Data::indexes.

 Debug:
    ::DEBUG bit 2 (==4 ) - debug log parsing and data storing
    ::DEBUG bit 3 (==8 ) - debug data referencing
    ::DEBUG bit 4 (==16) - debug initialization

=cut

#### Remarks

#TODO  Optimize data model for memory usage and access speed.
#      Idea:
#      $data->{$reftoindex}{$reftoaggr}{$key1val}{$key2val}...
#      this reduces hash depth twice and removes most uninformative
#      data out, but will get more time to search index and aggregation.
#      Using references as hash keys is weird, but should work fine.

use vars qw#@EXPORT_OK @functions $data @keyfields @sumfields @indexes
		@aggregations @associations#;
use Exporter;
use Utils qw#cmp_arrays printgroup#;
use Setup qw#dataclassen#;
use Carp qw#carp cluck croak confess#;
use Debug;

@Data::ISA = ('Exporter');

@EXPORT_OK = qw#addtolist debug dump exists grep path ptr setup sortindex
		store#;

$Data::NAME = 'Data';
$Data::VERSION = '0.1.0';
$Data::AUTHOR = 'qMax';
$Data::EMAIL = 'qmax@mail.ru';

# supported functions
@functions=qw(SUM COUNT MIN MAX AVG);

$data = {};

# list of available key fields
@keyfields=();
# list of available amount fields
@sumfields=();

# list of ALL requested indexes and groupings
# sorted in order of keyfields
@indexes=();

# list of ALL functions to use
#    [ $function , $field ]
@aggregations=();

# pairs [ \@function, \@index ]
# containing references to elements of @aggregations and @indexes
# each element forms single SELECT @index,$function GROUP BY @index
@associations=();

print $/, __FILE__, ": Saving data multi-dimensional" if ($::DEBUG);

=head2 STARTUP

Upon loading, reads list of available fields from parser module,
initializes @keyfields, dataclasses and relaxes.
Needs to be initialized from report module to setup indexes and
aggregations.

=cut

# actually Setup::dataclassen
our $dataclassen = dataclassen ();

for (keys %$dataclassen)
{
	my $field = $_;
	my $mainclass = $dataclassen->{$_}{'main'};
	
	if ($mainclass eq 'key')
	{
		push (@keyfields, $field);
	}
	elsif ($mainclass eq 'amount')
	{
		push (@sumfields, $field);
	}
	else
	{
		print STDERR $/, __FILE__, ": FATAL: Unknown mainclass '$mainclass'";
		die "\nBogus parser-module\n";
	}
}

if ($::DEBUG & 0x10)
{
	print $/, __FILE__, ': available data:';
	print "\n\tkeyfields:  ", printgroup (@keyfields);
	print "\n\taggregable: ", printgroup (@sumfields);
}

=head1 Functions

=cut

# sorts fields in proper order
# fields other then datafields or functions go to end
sub sortindex {
    my @i=@_;
    my @r=();
    foreach my $kf (@keyfields,@sumfields,@functions) {
	push @r,$kf if( grep($kf eq $_,@i) );
	}
    foreach my $kf (@i) {
	push @r,$kf unless( grep($kf eq $_,(@keyfields,@sumfields,@functions)) );
	}
    return @r;
}

# looks up $foo in @$list,
# returns reference to matching elem if found,
# or inserts $foo and returns it back.
sub addtolist
{
	my ($list, $arg) = @_;
	# $l is array-reference, as is $arg
	foreach my $l (@$list)
	{
		return $l if (cmp_arrays ($l, $arg));
	}

	push (@$list, $arg);
	return $arg;
}

=head2 setup @selections

Initializes supplemental structures.
Checks for valid fields and functions.
Should be callen prior to saving data.

 @selections is list of pairs: [ \@indexes, \@functions ]

Where @indexes is a list of combinations of keyfields going to be used,
and @functions is a list of [ $fun, $field ] to associate with
all that indexes.

Single selection equals to _several_ SQL statments, 
for each @fields in @indexes:
 SELECT @fields, @functions GROUP BY @fields;

Defining GROUP BY other then @index is not supported.

=cut

sub setup
{
    my (@selections)=@_;

    my $errors=0;

    foreach ( @selections ) {
	my ($inds,$aggs) = @$_;
	my @cur_inds=();
	my @cur_aggs=();

	# first sanity checks, empty index allowed
	unless( $inds ) {
	    print STDERR "\n$Data::NAME: FATAL: no indexes defined. Cannot aggregate data.";
	    $errors++;
	    last;
	    }
	unless( $aggs and @$aggs ) {
	    print STDERR "\n$Data::NAME: FATAL: no aggregations selected. No data to load. Nothing to do. No way to continue. Not implemented.";
	    $errors++;
	    last;
	    }
	
	foreach my $ind (@$inds) {
	    my @ind=();
	    foreach my $fld ( @$ind ) {
		unless( grep($_ eq $fld,@keyfields,@sumfields,'*') ) {
		    print STDERR "\n$Data::NAME: ERROR: $fld: requested data field is not available in log. Skipped";
		    next;
		    }
		if( grep($_ eq $fld,@sumfields) ) {
		    print STDERR "\n$Data::NAME: WARNING: $fld: uncountable data selected as keyfield. Expect memory overflow.";
		    }
		push @ind,$fld;
		}
	    @ind=sortindex(@ind);
	    # push if something left or nothing was
	    push @cur_inds, addtolist(\@indexes,\@ind) if( @ind or not @$ind);
	    }
	
	foreach my $agg (@$aggs) {
	    my ($fun,$fld)= @$agg;
	
	    unless( grep($_ eq $fun,@functions) ) {
		print STDERR "\n$Data::NAME: ERROR: $fun($fld): requested function not supported. Skipped.";
		next;
		}
	    unless( grep($_ eq $fld,@keyfields,@sumfields,'*') ) {
		print STDERR "\n$Data::NAME: ERROR: $fun($fld): requested data not available in log. Skipped";
		next;
		}

	    if( $fun eq 'COUNT' ) {
		unless( grep($_ eq $fld,(@keyfields,'*')) ) {
		    print STDERR "\n$Data::NAME: WARNING: $fun($fld): aggregable data found where countable expected. Expect bogus results.";
		    }
		}
	    else {
		unless( grep($_ eq $fld,@sumfields) ) {
		    print STDERR "\n$Data::NAME: FATAL: $fun($fld): requested function requires aggregable data.";
		    $errors++;
		    next;
		    }
		if( $fun eq 'AVG' ) {
		    # should be calculated before AVG
		    push @cur_aggs, addtolist(\@aggregations,['SUM',$fld]);
		    push @cur_aggs, addtolist(\@aggregations,['COUNT','*']);
		    }
		}
	    push @cur_aggs, addtolist(\@aggregations,$agg);
	    }

	# join cur_*
	my @cur_assocs=();
	foreach my $ind (@cur_inds) {
	    foreach my $aggr (@cur_aggs) {
		push @cur_assocs, [ $aggr, $ind ];
		}
	    }
	push @associations, @cur_assocs;
	}

    die "\n$Data::NAME: Dying because of $errors errors\n" if($errors);

    debug() if($::DEBUG&32);

    return 1;
}

# not best way to print it...
sub debug {
    print "\n$Data::NAME: Aggregating:";
    my @prevaggs=();
    foreach my $ind (@indexes) {
	my @aggs=();
	foreach my $ass (@associations) {
	    push @aggs,$ass->[0] if( $ass->[1] == $ind );
	    }
	unless( cmp_arrays(\@aggs,\@prevaggs) ) {
	    print "\n\t".join(' ', map("$_->[0]($_->[1])",@aggs))." BY:";
	    }
	print "\n\t".printgroup(@$ind) if( $ind );
	@prevaggs=@aggs;
	}
}

=head2 ptr \%KEY

Returns B<reference> to data, indexed by pairs (field=>value) in %KEY.

That is - either reference to scalar value of aggregation,
or reference to reference to hash for further indexing.

Thus, it should be accessed as $$ptr = value or $$ptr->{key} = value;

All referenced hashes and values are created (undefined or empty)
and inserted into tree automagically;

=cut

# Automagic works, because perl allows to dereference undef value
# as a hash, or as a scalar, untill either hash or scalar get modified.

sub ptr
{
    my $key=shift;
    my $dataref=\$data;
    print "\nPTR " if $::DEBUG&8;
    foreach my $k (sortindex(keys %$key)) {
	my $kv = $key->{$k};
	print (defined $$dataref ? $$dataref : 'NULL' ) if $::DEBUG&8;
	$dataref=\$$dataref->{$k}{$kv};
	print "[$k:$kv]=" if $::DEBUG&8;
	}

    print (defined $$dataref ? $$dataref : 'NULL' ) if $::DEBUG&8;
    return $dataref;
}

=head2 exists \%KEY

Works like ptr, but returns undef if some data is missed
in data tree. That is - does not create nonexistant hashes.

=cut

sub exists
{
    my $key=shift;
    my $dataref=\$data;
    print "\nCHKPTR " if $::DEBUG&8;
    foreach my $k (sortindex(keys %$key)) {
	my $kv = $key->{$k};
	print (defined $$dataref ? $$dataref : 'NULL' ) if $::DEBUG&8;
	$dataref=\$$dataref->{$k}{$kv};
	print "[$k:$kv]=" if $::DEBUG&8;
	last unless(defined $$dataref);
	}
    print (defined $$dataref ? $$dataref : 'NULL' ) if $::DEBUG&8;
    return undef unless( defined $$dataref );
    return $dataref;
}

=head2 grep %KEY

Returns all available keys, matching given %KEY, with
undef keyvalues replaced with each available.

Note, that %KEY may be incomplete (pointing to deeper hash)

Returns empty list if such keys dont exist in database.

=cut

#
# Algo is clean:
# For each field in index, append it to all already greped keys,
# either with given value or with all avaliable values.
# If undef is in middle of index, some keys got with 'all available'
# values will be thrown away.
#

sub grep
{
    my %matchkey=@_;
    my @keys=({});

    foreach my $kf( sortindex(keys %matchkey) ) {
	my $kv = $matchkey{$kf};
	my @tmp=@keys; @keys=();
	while( my $key = shift @tmp ) {

	    # find out ALL available values
	    my $dataref=Data::exists($key);
	    # there's no such key at all or no deeper hash
	    return () unless( defined $dataref and ref($$dataref) eq 'HASH');

	    $dataref = $$dataref->{$kf};
	    my @values = keys %$dataref;
	    if( defined $kv ) {
		# check if given value present
		if( grep( $_ eq $kv, @values ) ) {
		    $key->{$kf}=$kv;
		    push @keys,{%$key};
		    }
		}
	    else {
		# add all taken values
		foreach my $kv ( @values ) {
		    $key->{$kf}=$kv;
		    push @keys,{%$key};
		    }
		}
	    }
	}
    return @keys;
}


=head2 path \%KEY

Debugging: draws key path of fields in %KEY in form [key1:val1][key2:val2]...

In the same order as actually stored.

=cut

sub path
{
    my $key=shift;
    my $path="";
    foreach my $f (sortindex(keys %$key)) {
	$path.="[$f:$key->{$f}]" if ( defined $key->{$f} );
	}
    return $path;
}

=head2 store \%DATARECORD

Stores data in database, performs all aggregations.

=cut

sub store
{
	my $hash = shift;

	# list of ALL keys indexed by idx-refs
	my $keys = {};
	foreach my $ind (@indexes)
	{
		$keys->{$ind} = {};
		$keys->{$ind}{$_} = $hash->{$_} for (@$ind);
	}

	print $/, __FILE__, ": store: ",
		join (' ', map { "$_=" . $hash->{$_} . ';' } (keys %$hash))
		if ($::DEBUG & 4);

	foreach my $assoc (@associations)
	{
		my ($aggr, $idx, $grp) = @$assoc;
		my ($fun, $fld) = @$aggr;
		my $key = $keys->{$idx};
	
		print $/, __FILE__, ": UPD: ", path ({ %$key,$fun=>$fld })
			if ($::DEBUG & 4);
		
		my $dataref = ptr ({%$key, $fun => $fld});
		my $value = $hash->{$fld} unless ($fld eq '*');

		if ($fun eq 'SUM')
		{
			$$dataref += $value;
			print " += $value" if ($::DEBUG & 4);
		}
		elsif ($fun eq 'AVG')
		{
			my $sumref = ptr ({%$key, 'SUM' => $fld});
			my $cntref = ptr ({%$key, 'COUNT' => '*'});
			$$dataref = $$cntref ? int ($$sumref / $$cntref) : 0;
			print " = $$dataref" if ($::DEBUG & 4);
		}
		elsif ($fun eq 'MIN')
		{
			if (!defined ($$dataref) or $value < $$dataref)
			{
				$$dataref = $value;
			}
			print " = $value" if ($::DEBUG & 4);
		}
		elsif ($fun eq 'MAX')
		{
			if (!defined($$dataref) or $value > $$dataref)
			{
				$$dataref = $value;
			}
			print " = $value" if ($::DEBUG & 4);
		}
		elsif ($fun eq 'COUNT' and $fld eq '*')
		{
			$$dataref++;
			print "++" if ($::DEBUG & 4);
		}
		elsif ($fun eq 'COUNT')
		{
			my $countedref = ptr ({%$key, 'COUNTED' => $fld});
			$$countedref = [] unless ($countedref);
			unless (grep { $_ eq $value } (@$$countedref))
			{
				push (@$$countedref, $value);
				$$dataref++;
				print "++ ($value)" if ($::DEBUG & 4);
			}
		}
	}
}

=head2 dump $DATAREF, \%KEY

Dumps referenced data recursively.
Shows actual data stored, not the data, requested to store.

=cut

sub dump
{
    my ($key)=@_;
    $key={} unless($key);

    my $ref=ptr($key);

    if( ref($$ref) eq 'HASH' ) {
	while( my($kf,$kfv)=each %$$ref) {
	    if( ref($kfv) eq 'HASH' ) {
		foreach my $kv ( keys %$kfv ) {
		    Data::dump({%$key,$kf=>$kv});
		    }
		}
	    else {
		# some bogus stuff
		print "\n".path($key);
		print "[$kf] = $kfv";
		}
	    }
	return;
	}
#    if( ref($$ref) eq 'HASH' ) {
#	foreach my $k (keys %$$ref) {
#	    my($kf,$kv) = ($k=~/^(.*?):(.*)$/);
#	    Data::dump({%$key,$kf=>$kv});
#	    }
#	return;
#	}

    # stored data
    print "\n".path($key);

    # if there's another structure
    if( ref($ref) eq 'REF' ) {
	print " -> ";
	$ref=$$ref;
	}
    else { print " = "; }

    if( ref($ref) eq 'SCALAR' ) {
	if( $$ref ) { print "'$$ref'"; } else { print "NULL"; }
	return;
	}
    if( ref($ref) eq 'ARRAY' ) {
	print "(".join(' ',@$ref).")";
	return;
	}
    if( ref($ref) eq 'HASH' ) {
	print "(".join(' ',map("$_=>$ref->{$_}",keys %$ref) );
	return;
	}

    print "BOGUS ref $ref";
}

=head1 AUTHOR

QMax E<lt>qmax@mail.ruE<gt>

=cut

1;
